/* This file contains all the functions required for the standalone
   ip_nat module.

   These are not required by the compatibility layer.
*/

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#ifdef MODULE
#define EXPORT_SYMTAB
#endif
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <linux/spinlock.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_nat_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_nat_lock)

#include "conntrack/struct_reserve.h"
#include "NAT/ip_nat.h"
#include "NAT/ip_nat_rule.h"
#include "NAT/ip_nat_protocol.h"
#include "NAT/ip_nat_core.h"
#include "NAT/ip_nat_helper.h"
#include "NAT/ip_nat_expect.h"
#include "conntrack/ip_conntrack_core.h"
#include "common/listhelp.h"

static inline unsigned int
alloc_null_binding(struct ip_conntrack *conntrack,
		   struct ip_nat_info *info,
		   unsigned int hooknum)
{
	/* Force range to this IP; let proto decide mapping for
	   per-proto parts (hence not IP_NAT_RANGE_PROTO_SPECIFIED).
	   Use reply in case it's already been mangled (eg local packet).
	*/
	u_int32_t ip
		= (HOOK2MANIP(hooknum) == IP_NAT_MANIP_SRC
		   ? conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.dst.ip
		   : conntrack->tuplehash[IP_CT_DIR_REPLY].tuple.src.ip);
	struct ip_nat_multi_range mr
		= { 1, { { IP_NAT_RANGE_MAP_IPS, ip, ip, { 0 }, { 0 } } } };

	return ip_nat_setup_info(conntrack, info, &mr, hooknum);
}

/* All zero name for locally-generated packets incoming `interface' */
static char local_name[IFNAMSIZ] = { 0 };

static unsigned int
init_conntrack(struct ip_conntrack *conntrack,
	       struct ip_nat_info *info,
	       unsigned int hooknum,
	       const char *ifname,
	       struct sk_buff *skb)
{
	struct ip_nat_rule *rule;
	unsigned int ret;

	/* Protects the NAT private region of the conntrack. */
	MUST_BE_WRITE_LOCKED(&ip_nat_lock);
	rule = __ip_nat_rule_get(&conntrack->tuplehash[IP_CT_DIR_ORIGINAL]
				 .tuple, ifname,
				 hooknum);

	/* If there's a rule which matches, set up the translation. */
	if (rule) {
		if (rule->mtype)
			ret = rule->mtype->setup(conntrack, info, hooknum, skb,
						 &rule->mr,
						 rule->mopts);
		else
			ret = ip_nat_setup_info(conntrack, info, &rule->mr,
						hooknum);

		ip_nat_rule_put(rule);
	} else {
		/* Set up null or pseudo-null binding. */
		ret = alloc_null_binding(conntrack, info, hooknum);
	}
	return ret;
}

#define HOOKNAME(hooknum) ((hooknum) == NF_IP_POST_ROUTING ? "POST_ROUTING"  \
			   : ((hooknum) == NF_IP_PRE_ROUTING ? "PRE_ROUTING" \
			      : ((hooknum) == NF_IP_LOCAL_OUT ? "LOCAL_OUT"  \
				 : "*ERROR*")))

#ifdef DEBUG_CONNTRACK
void DUMP_MANIP(unsigned int num_manips,
		const struct ip_nat_info_manip *manips)
{
	unsigned int i;

	for (i=0; i < num_manips; i++) {
		DEBUGP("manip %u: dir %s hook %s manip %s"
		       " ip %u.%u.%u.%u port %u\n",
		       i,
		       manips[i].direction == IP_CT_DIR_ORIGINAL ? "ORIG"
		       : (manips[i].direction == IP_CT_DIR_REPLY
			  ? "REPLY" : "*ERROR*"),
		       HOOKNAME(manips[i].hooknum),
		       manips[i].maniptype == IP_NAT_MANIP_SRC ? "SRC"
		       : (manips[i].maniptype == IP_NAT_MANIP_DST
			  ? "DST" : "*ERROR*"),
		       IP_PARTS(manips[i].manip.ip),
		       ntohs(manips[i].manip.u.all));
	}
}
#else
#define DUMP_MANIP(manips)
#endif

static unsigned int
ip_nat_fn(unsigned int hooknum,
	  struct sk_buff **pskb,
	  const struct net_device *in,
	  const struct net_device *out,
	  int (*okfn)(struct sk_buff *))
{
	struct ip_conntrack_tuple_hash *h;
	enum ip_conntrack_info ctinfo;
	struct ip_nat_info *info;
	unsigned int ret = NF_ACCEPT;
	/* maniptype == SRC for postrouting. */
	enum ip_nat_manip_type maniptype = HOOK2MANIP(hooknum);

	/* We never see framents: conntrack defrags on pre-routing and
	   local-out, and ip_nat_out protects post-routing. */
	IP_CONNTRACK_ASSERT(!((*pskb)->nh.iph->frag_off
			      & __constant_htons(IP_MF|IP_OFFSET)));

	/* FIXME: One day, fill in properly. --RR */
	(*pskb)->nfcache |= NFC_UNKNOWN | NFC_ALTERED;

	/* If we had a hardware checksum before, it's now invalid */
	if ((*pskb)->pkt_type != PACKET_LOOPBACK)
		(*pskb)->ip_summed = CHECKSUM_NONE;

	/* If we have partially mangled a packet, and ip_conntrack_get
	   does a lookup, it could fail.  Doing refcnts on conntracks
	   will clean this up. */
	if (hooknum == NF_IP_POST_ROUTING) {
		if (!ip_conntrack_easy_get(*pskb, &h, &ctinfo)) {
			struct ip_conntrack_tuple tuple, repl_tuple;
			/* FIXME: ICMP errors here --RR */

			/* Look up backwards: reply should match. */
			if (!get_tuple((*pskb)->nh.iph, (*pskb)->len, &tuple,
				       find_proto((*pskb)->nh.iph->protocol))
			    || !invert_tuplepr(&repl_tuple, &tuple)) {
				ctinfo = IP_CT_INVALID;
			} else {
				ctinfo = ip_conntrack_hard_get(*pskb,
							       &repl_tuple,
							       &h);
			}

			/* We want the other half */
			if (ctinfo != IP_CT_INVALID)
				h = &h->ctrack->tuplehash[!DIRECTION(h)];
		}
	}
	else ctinfo = ip_conntrack_get(*pskb, &h);

	switch (ctinfo) {
	case IP_CT_INVALID:
		ret = NF_ACCEPT;
		goto out;

	case IP_CT_RELATED:
		if ((*pskb)->nh.iph->protocol == IPPROTO_ICMP) {
			icmp_reply_translation(*pskb, h->ctrack, hooknum,
					       DIRECTION(h));
			ret = NF_ACCEPT;
			goto release;
		}
		/* Fall thru... */
	case IP_CT_NEW:
		info = STRUCT_RESERVED_PTR(ip_conntrack, h->ctrack,
					   &ip_nat_offset);

		if (!info) {
			ret = NF_ACCEPT;
			goto release;
		}

		WRITE_LOCK(&ip_nat_lock);
		/* Seen it before?  This can happen for loopback, retrans,
		   or local packets.. */
		if (!(info->initialized & (1 << maniptype))) {
			const char *ifname;

			if (hooknum == NF_IP_LOCAL_OUT)
				ifname = local_name;
			else if (hooknum == NF_IP_PRE_ROUTING)
				ifname = in->name;
			else {
				IP_CONNTRACK_ASSERT(hooknum
						    == NF_IP_POST_ROUTING);
				ifname = out->name;
			}

#if 0
			DEBUGP("Not initialized for maniptype %s hook %s\n",
			       maniptype == IP_NAT_MANIP_SRC ? "SRC" : "DST",
			       hooknum == NF_IP_POST_ROUTING ? "POST_ROUTING"
			       : (hooknum == NF_IP_PRE_ROUTING ? "PRE_ROUTING"
				  : "LOCAL_OUT"));
#endif
			ret = init_conntrack(h->ctrack, info, hooknum,
					     ifname, *pskb);

			if (ret == NF_ACCEPT) {
				if (info->initialized) {
					IP_CONNTRACK_ASSERT(info->bysource
							    .conntrack);
					replace_in_hashes(h->ctrack, info);
				} else {
					place_in_hashes(h->ctrack, info);
				}
				info->initialized |= (1 << maniptype);
#if 0
				DEBUGP("After %s init:\n",
				       maniptype == IP_NAT_MANIP_SRC ? "SRC"
				       : "DST");
				DUMP_MANIP(info->num_manips, info->manips);
#endif
			}
			else DEBUGP("init_conntrack returned %u\n", ret);
		}
		WRITE_UNLOCK(&ip_nat_lock);
		if (ret != NF_ACCEPT)
			goto release;
		break;

	default:
		/* ESTABLISHED / ESTABLISHED_SPECIFIES_RELATED */
		info = STRUCT_RESERVED_PTR(ip_conntrack, h->ctrack,
					   &ip_nat_offset);

		if (!info) {
			ret = NF_ACCEPT;
			goto release;
		}
	}

	IP_CONNTRACK_ASSERT(info);
	ret = do_bindings(h, ctinfo, info, hooknum, pskb);
 release:
	ip_conntrack_put(h->ctrack);
 out:
	return ret;
}

static unsigned int
ip_nat_out(unsigned int hooknum,
	   struct sk_buff **pskb,
	   const struct net_device *in,
	   const struct net_device *out,
	   int (*okfn)(struct sk_buff *))
{
	/* We can hit fragment here; forwarded packets get
	   defragmented by connection tracking coming in, then
	   fragmented (grr) by the forward code.  We have to redefrag
	   them in that case.

	   FIXME: If they've been partially mangled (double NAT), we
	   won't find a connection for them.

	   I'm starting to have nightmares about fragments.
	*/

	if ((*pskb)->nh.iph->frag_off & __constant_htons(IP_MF|IP_OFFSET)) {
#ifdef CONFIG_NETFILTER_DEBUG
		unsigned int olddebug = (*pskb)->nf_debug;
#endif
		DEBUGP("ip_nat_out: gathering fragments on %p\n", *pskb);
		*pskb = ip_defrag(*pskb);
		if (!*pskb)
			return NF_STOLEN;
		ip_send_check((*pskb)->nh.iph);
		(*pskb)->nfcache |= NFC_ALTERED;
#ifdef CONFIG_NETFILTER_DEBUG
		/* Packet path as if nothing had happened. */
		(*pskb)->nf_debug = olddebug;
#endif
		DEBUGP("ip_nat_out: gathering successful\n");
	}

	return ip_nat_fn(hooknum, pskb, in, out, okfn);
}

/* Packet filtering is priority 0: we must be after connection
   tracking (-200). */

/* Before packet filtering, change destination */
static struct nf_hook_ops ip_nat_in_ops
= { { NULL, NULL }, ip_nat_fn, NULL, PF_INET, NF_IP_PRE_ROUTING, -100 };
/* After packet filtering, change source */
static struct nf_hook_ops ip_nat_out_ops
= { { NULL, NULL }, ip_nat_out, NULL, PF_INET, NF_IP_POST_ROUTING, 100 };
/* Before packet filtering, change destination */
static struct nf_hook_ops ip_nat_local_out_ops
= { { NULL, NULL }, ip_nat_fn, NULL, PF_INET, NF_IP_LOCAL_OUT, -100 };

/* Protocol registration. */
int ip_nat_protocol_register(struct ip_nat_protocol *proto)
{
	int ret = 0;
	struct list_head *i;

	WRITE_LOCK(&ip_nat_lock);
	for (i = protos.next; i != &protos; i = i->next) {
		if (((struct ip_nat_protocol *)i)->protonum
		    == proto->protonum) {
			ret = -EBUSY;
			goto out;
		}
	}

	list_prepend(&protos, proto);
	MOD_INC_USE_COUNT;

 out:
	WRITE_UNLOCK(&ip_nat_lock);
	return ret;
}

/* Noone stores the protocol anywhere; simply delete it. */
void ip_nat_protocol_unregister(struct ip_nat_protocol *proto)
{
	WRITE_LOCK(&ip_nat_lock);
	LIST_DELETE(&protos, proto);
	WRITE_UNLOCK(&ip_nat_lock);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,43)
	/* Someone could be still looking at the proto in a bh.  Wait
           for them. */
	synchronize_bh();
#endif

	MOD_DEC_USE_COUNT;
}

static int init_or_cleanup(int init)
{
	int ret = 0;

	if (!init) goto cleanup;

	ret = ip_nat_rule_init();
	if (ret < 0) {
		printk("ip_nat_init: can't setup rules.\n");
		goto cleanup_nothing;
	}
	ret = ip_nat_init();
	if (ret < 0) {
		printk("ip_nat_init: can't setup rules.\n");
		goto cleanup_rule_init;
	}

	ret = nf_register_hook(&ip_nat_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register in hook.\n");
		goto cleanup_nat;
	}
	ret = nf_register_hook(&ip_nat_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register out hook.\n");
		goto cleanup_inops;
	}
	ret = nf_register_hook(&ip_nat_local_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local out hook.\n");
		goto cleanup_outops;
	}
	__MOD_INC_USE_COUNT(ip_conntrack_module);
	return ret;

 cleanup:
	__MOD_DEC_USE_COUNT(ip_conntrack_module);
	nf_unregister_hook(&ip_nat_local_out_ops);
 cleanup_outops:
	nf_unregister_hook(&ip_nat_out_ops);
 cleanup_inops:
	nf_unregister_hook(&ip_nat_in_ops);
 cleanup_nat:
	ip_nat_cleanup();
 cleanup_rule_init:
	ip_nat_rule_cleanup();
 cleanup_nothing:
	MUST_BE_READ_WRITE_UNLOCKED(&ip_nat_lock);
	MUST_BE_READ_WRITE_UNLOCKED(&ip_conntrack_lock);
	return ret;
}

int __init ip_init_nat(void)
{
	return init_or_cleanup(1);
}

void __exit ip_nat_exit(void)
{
	init_or_cleanup(0);
}

module_init(ip_init_nat);
module_exit(ip_nat_exit);

EXPORT_SYMBOL(ip_nat_setup_info);
EXPORT_SYMBOL(ip_nat_register_mapping_type);
EXPORT_SYMBOL(ip_nat_unregister_mapping_type);
EXPORT_SYMBOL(ip_nat_helper_register);
EXPORT_SYMBOL(ip_nat_helper_unregister);
EXPORT_SYMBOL(ip_nat_expect);
EXPORT_SYMBOL(ip_nat_cheat_check);

